#include <sys/types.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hw_md.h"
#include "files.h"

extern int sh2_app_length;
extern char sh2_app_start[];

#define MAX_ENTRIES 256

typedef struct
{
    int offset;
    int length;
    char flags;
    char name[247];
} entry_t;

entry_t gDirTable[MAX_ENTRIES];

short gCount = 0;
short gFirst = 0;
short gCurrent = 0;


void delay(int count)
{
    int t = GET_TICKS + count;
    while (t > GET_TICKS) ;
}

int getCurrDir(void)
{
    int r = 0;
    int i = 0;

    while (!r)
    {
        r = next_dir_entry();
        if (!r && memcmp(global_vars->DENTRY_NAME, ".", 2))
        {
            gDirTable[i].offset = global_vars->DENTRY_OFFSET;
            gDirTable[i].length = global_vars->DENTRY_LENGTH;
            gDirTable[i].flags = global_vars->DENTRY_FLAGS >> 8;
            strcpy(gDirTable[i].name, global_vars->DENTRY_NAME);
            i++;
        }
    }

    return i;
}

int main(void)
{
    char *temp = (char *)0x0C0000;
    int i, j, x;
    unsigned short buttons, changed, previous = 0;

    strcpy(temp, "CD32X MOD Player");
    switch_banks();
    do_md_cmd4(MD_CMD_PUT_STR, 0x200000, TEXT_WHITE, 12, 1);

    memcpy(temp, &sh2_app_start[8], sh2_app_length - 8);
    switch_banks();
    x = do_md_cmd2(MD_CMD_INIT_32X, 0x200000, sh2_app_length);
    switch (x)
    {
        case -1:
            strcpy(temp, "No 32X support in hw_md.s file!");
            switch_banks();
            do_md_cmd4(MD_CMD_PUT_STR, 0x200000, TEXT_RED, 12, 14);
            break;
        case -2:
            strcpy(temp, "32X not detected!");
            switch_banks();
            do_md_cmd4(MD_CMD_PUT_STR, 0x200000, TEXT_RED, 12, 14);
            break;
        case -3:
            strcpy(temp, "32X SDRAM error!");
            switch_banks();
            do_md_cmd4(MD_CMD_PUT_STR, 0x200000, TEXT_RED, 12, 14);
            break;
        default:
            strcpy(temp, "32X started...");
            switch_banks();
            do_md_cmd4(MD_CMD_PUT_STR, 0x200000, TEXT_GREEN, 13, 3);
            delay(120);
            strcpy(temp, "              ");
            switch_banks();
            do_md_cmd4(MD_CMD_PUT_STR, 0x200000, TEXT_GREEN, 13, 3);
            break;
    }

    init_cd(); // init CD
    set_cwd("/"); // set current directory to root
    gCount = getCurrDir();

    while (1)
    {
        char *p = temp;

        // update display with directory entries
        memset(p, 0, 20*64*2);
        if (gCount)
            for (i = 0; i < 20; i++)
            {
                if ((gFirst + i) >= gCount)
                    break;

                p = &temp[i*64*2 + 2]; // left justified
                if (gDirTable[gFirst + i].flags & 2)
                {
                    // directory
                    *p++ = ((gFirst + i) == gCurrent) ? TEXT_WHITE >> 8 : TEXT_GREEN >> 8;
                    *p++ = '[' - 0x20;
                    for (j = 0; j < 36; j++)
                    {
                        if (gDirTable[gFirst + i].name[j] == '\0')
                            break;
                        *p++ = ((gFirst + i) == gCurrent) ? TEXT_WHITE >> 8 : TEXT_GREEN >> 8;
                        *p++ = gDirTable[gFirst + i].name[j] - 0x20;
                    }
                    *p++ = ((gFirst + i) == gCurrent) ? TEXT_WHITE >> 8 : TEXT_GREEN >> 8;
                    *p++ = ']' - 0x20;
                }
                else
                {
                    for (j = 0; j < 38; j++)
                    {
                        if (gDirTable[gFirst + i].name[j] == '\0')
                            break;
                        *p++ = ((gFirst + i) == gCurrent) ? TEXT_WHITE >> 8 : TEXT_GREEN >> 8;
                        *p++ = gDirTable[gFirst + i].name[j] - 0x20;
                    }
                }
            }
        switch_banks();
        delay(1);
        do_md_cmd3(MD_CMD_COPY_VRAM, 0xC000 + 3*64*2, 0x200000, 20*64); // update plane A name table (20 lines of text)

        delay(5);
        buttons = GET_PAD(0);

        changed = (buttons & SEGA_CTRL_BUTTONS) ^ previous;

        if (buttons & SEGA_CTRL_DOWN)
        {
            // DOWN pressed
            gCurrent++;
            if ((gCurrent - gFirst) >= 20)
                gFirst = gCurrent;

            if (gCurrent >= gCount)
                gFirst = gCurrent = 0; // wrap around to first entry
        }

        if (buttons & SEGA_CTRL_UP)
        {
            // UP pressed
            gCurrent--;
            if (gCurrent < gFirst)
            {
                gFirst = gCurrent - 19;
                if (gFirst < 0)
                    gFirst = 0;
            }

            if (gCurrent < 0)
            {
                gCurrent = gCount - 1; // wrap around to last entry
                gFirst = gCurrent - 19;
                if (gFirst < 0)
                    gFirst = 0;
            }
        }

        if (!changed)
            continue; // no change in buttons
        previous = buttons & SEGA_CTRL_BUTTONS;

        if (changed & SEGA_CTRL_A)
        {
            if (!(buttons & SEGA_CTRL_A))
            {
                // A just released
                if (gDirTable[gCurrent].flags & 2)
                {
                    // directory
                    global_vars->CWD_OFFSET = gDirTable[gCurrent].offset;
                    global_vars->CWD_LENGTH = gDirTable[gCurrent].length;
                    first_dir_sec();
                    gCount = getCurrDir();
                    gFirst = gCurrent = 0;
                }
                else
                {
                    // check if not too big
                    if (gDirTable[gCurrent].length > (128*1024))
                    {
                        sprintf(temp, "File too big! %d bytes", gDirTable[gCurrent].length);
                        j = strlen(temp) / 2; // because we are going to switch banks...
                        switch_banks();
                        do_md_cmd4(MD_CMD_PUT_STR, 0x200000, TEXT_RED, 20 - j, 25);
                        delay(120);
                        strcpy(temp, "                                      ");
                        switch_banks();
                        do_md_cmd4(MD_CMD_PUT_STR, 0x200000, TEXT_GREEN, 1, 25);
                        continue;
                    }
                    else
                    {
                        sprintf(temp, "Loading %s: %d words", gDirTable[gCurrent].name, (gDirTable[gCurrent].length + 1) >> 1);
                        j = strlen(temp) / 2; // because we are going to switch banks...
                        switch_banks();
                        do_md_cmd4(MD_CMD_PUT_STR, 0x200000, TEXT_RED, 20 - j, 25);
                        delay(120);
                        strcpy(temp, "                                      ");
                        switch_banks();
                        do_md_cmd4(MD_CMD_PUT_STR, 0x200000, TEXT_GREEN, 1, 25);
                    }
                    // check if supported MOD
                    load_file(gDirTable[gCurrent].name, temp);
                    if (memcmp(&temp[0x438], "M.K.", 4) && memcmp(&temp[0x438], "4CHN", 4) && memcmp(&temp[0x438], "6CHN", 4) && memcmp(&temp[0x438], "8CHN", 4))
                    {
                        strcpy(temp, "Not a supported MOD file!");
                        switch_banks();
                        do_md_cmd4(MD_CMD_PUT_STR, 0x200000, TEXT_RED, 8, 25);
                        delay(120);
                        strcpy(temp, "                                      ");
                        switch_banks();
                        do_md_cmd4(MD_CMD_PUT_STR, 0x200000, TEXT_GREEN, 1, 25);
                        continue;
                    }
                    // send to 32X and start player
                    if (!x)
                    {
#if 0
                        // set COMM0 to "get ready to dma mod file, then play it" command
                        do_md_cmd3(MD_CMD_SET_COMM32X, 0, 4, 0x00010000);
                        // dma the data
                        switch_banks();
                        j = do_md_cmd2(MD_CMD_DMA_TO_32X, 0x200000, (gDirTable[gCurrent].length + 1) >> 1);
#else
                        // set COMM0 to "get ready to xfer mod file, then play it" command
                        do_md_cmd3(MD_CMD_SET_COMM32X, 0, 4, 0x00010000);
                        // xfer the data
                        switch_banks();
                        j = do_md_cmd2(MD_CMD_CPY_TO_32X, 0x200000, (gDirTable[gCurrent].length + 1) >> 1);
#endif
                        if (j < 0)
                        {
                            sprintf(temp, " Error: %d", j);
                            switch_banks();
                            do_md_cmd4(MD_CMD_PUT_STR, 0x200000, TEXT_RED, 13, 25);
                            delay(120);
                            strcpy(temp, "                                      ");
                            switch_banks();
                            do_md_cmd4(MD_CMD_PUT_STR, 0x200000, TEXT_GREEN, 1, 25);
                        }
                        else
                        {
                            j = do_md_cmd2(MD_CMD_GET_COMM32X, 0, 4);
                            while (j != 0x00000001)
                                j = do_md_cmd2(MD_CMD_GET_COMM32X, 0, 4);
                            strcpy(temp, "Playing MOD...");
                            switch_banks();
                            do_md_cmd4(MD_CMD_PUT_STR, 0x200000, TEXT_RED, 13, 25);
                            j = do_md_cmd2(MD_CMD_GET_COMM32X, 0, 4);
                            while (j)
                                j = do_md_cmd2(MD_CMD_GET_COMM32X, 0, 4);
                            strcpy(temp, "                                      ");
                            switch_banks();
                            do_md_cmd4(MD_CMD_PUT_STR, 0x200000, TEXT_GREEN, 1, 25);
                        }
                    }
                    else
                    {
                        strcpy(temp, "Can't play MOD without 32X!");
                        switch_banks();
                        do_md_cmd4(MD_CMD_PUT_STR, 0x200000, TEXT_RED, 7, 25);
                        delay(120);
                        strcpy(temp, "                                      ");
                        switch_banks();
                        do_md_cmd4(MD_CMD_PUT_STR, 0x200000, TEXT_GREEN, 1, 25);
                        continue;
                    }
                }
            }
        }
    }

    return 0;
}

